#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "../../lib/ucos2.h"
#include "../../lib/audio.h"

#include "../modplay/modplay.h"

#include "a320_snd.h"

typedef void waveout_inst;

typedef struct {
	unsigned int sample_rate;
	unsigned short format;
	unsigned char channel;
	unsigned char volume;
} waveout_args;

#define min(a,b) (a<b ? a : b)

waveout_inst *_audio_music = NULL;
waveout_args _audioargs = { A320_SAMPLE_RATE, 16, 1, 20 };
  
//----------------------------------------------------------------------------------------------
#define  TASK_START_PRIO 5
#define  TASK_STK_SIZE   1024                      // Size of each task's stacks (# of WORDs)
#define  NO_TASKS        1                         // Number of identical tasks 

OS_STK   TaskStk[NO_TASKS][TASK_STK_SIZE];          // Tasks stacks 
OS_STK   TaskStartStk[TASK_STK_SIZE];
char     TaskData[NO_TASKS];                        // Parameters to pass to each task

static volatile int a320_sound_pausei;
static volatile int a320_sound_mod_stop;
static volatile int a320_sound_thread_exit;			  // Flag to end a320_new_sound_play() thread

static signed short a320_sound_bufmod[A320_SNDLENGTH];  // Sound buffer for mod
signed short a320_sound_mod[A320_SNDLENGTH];  // Sound buffer for mod

int a320_sound_level;
a320_channel_t a320_channel[A320_MIXCHANNELS];
MODFILE mod;

void _mixerPlay(void *none) {
  MODFILE *modthread = &mod;
  uint32_t left, data, bcl, chan;
  int16_t dataValue, *pRaw , *pDest;
  
  do { 
    if (a320_sound_pausei == 0) {
      if (a320_sound_mod_stop == 0) {
        modthread->mixingbuf = (void*) a320_sound_mod;
        modthread->mixingbuflen = A320_SNDLENGTH;
        MODFILE_Player(modthread);
        
        for (bcl = 0; bcl < A320_SNDLENGTH; bcl++) {
          a320_sound_bufmod[bcl] = a320_sound_mod[bcl];
        }
      }
      
      for (chan = 0; chan < A320_MIXCHANNELS; chan++) {
        if (a320_channel[chan].loop != 0) {
        	pRaw = (a320_channel[chan].buf + a320_channel[chan].index);
          pDest = a320_sound_bufmod;
          if (a320_channel[chan].index < a320_channel[chan].len) {
            left = (a320_channel[chan].len - a320_channel[chan].index);
            data = min(left, A320_SNDLENGTH);

            // If bits = 16
            //data = data / 2;
            uint32_t i;
            for (i = 0; i < data; i++, pRaw++, pDest++)
            {
              dataValue = (*pRaw)>>3;  // change sound volume (12.5%)
              if (dataValue != 0)
              {
                if (dataValue >= 0 && *pDest >= 0)
                {
                  *pDest = dataValue + *pDest - (dataValue * *pDest / 32768);
                }
                else if (dataValue < 0 && *pDest < 0)
                {
                  *pDest = dataValue + *pDest + (dataValue * *pDest / 32768);
                }
                else // different sign
                {
                  *pDest = dataValue + *pDest;
                }
              }
            }
            a320_channel[chan].index += A320_SNDLENGTH;
          }
          else {
            a320_channel[chan].loop--;
            if (a320_channel[chan].loop > 0) 
              a320_sound_endchannel(chan);
            else
              a320_channel[chan].index = 0;
          }
            
        }
      }
      waveout_write(_audio_music, a320_sound_bufmod, A320_SNDLENGTH); 
		} 
    else
      OSTimeDly(1);
	} while(a320_sound_thread_exit == 0);							                    // Until the end of the sound thread
  
	a320_sound_thread_exit=0;

  OSTaskDel(TASK_START_PRIO);
}

void a320_sound_endchannel(uint8_t chan)
{
	a320_channel[chan].len = 0;
	a320_channel[chan].index = 0;
	a320_channel[chan].loop = 0;
	a320_channel[chan].buf = NULL;
	a320_channel[chan].sndbuf = NULL;
}

void a320_sound_pause(int yes) 
{ 
  a320_sound_pausei = yes; 
  a320_sound_mod_stop = yes;
}

void a320_sound_thread_mute(void)
{
  uint8_t chan;
  
	memset(a320_sound_bufmod,0,A320_SNDLENGTH);
  
  // Desactivate all wav channel
  for (chan = 0; chan < A320_MIXCHANNELS; chan++) {
    a320_sound_endchannel(chan);
  }
}

void a320_sound_thread_stop(void)
{
  a320_sound_pausei = 1;
	a320_sound_thread_exit=1;
	
	a320_sound_thread_mute();
}

void a320_sound_init(void) {
  a320_sound_thread_exit=0;
  a320_sound_pausei = 1;
  
  _audio_music = waveout_open(&_audioargs);

  a320_sound_level = VOLUME_INIT;
  waveout_set_volume(a320_sound_level);

  a320_sound_thread_mute();

  MODFILE_Init(&mod);

  OSTaskCreate(_mixerPlay, (void *) 0, (void *)&TaskStartStk[TASK_STK_SIZE - 1], TASK_START_PRIO);
}

void a320_sound_stop(void) {
  a320_sound_thread_stop();

  while (a320_sound_thread_exit);  // to avoid pb with waveout_close -> can hang ??

  waveout_close(_audio_music);
  
  MODFILE_Stop(&mod);
  MODFILE_Free(&mod);
}

void a320_sound_increase(void)
{
	if (a320_sound_level < VOLUME_MAX) a320_sound_level += VOLUME_STEP;
	if (a320_sound_level > VOLUME_MAX) a320_sound_level = VOLUME_MAX;

  waveout_set_volume(a320_sound_level);
}

void a320_sound_decrease(void)
{
	if (a320_sound_level > 0) a320_sound_level -= VOLUME_STEP;
	if (a320_sound_level < 0) a320_sound_level = 0;

  waveout_set_volume(a320_sound_level);
}

void a320_sound_modstop(void) {
  //a320_sound_pausei = 1;
  a320_sound_mod_stop = 1;
  memset(a320_sound_bufmod,0,A320_SNDLENGTH);
  MODFILE_Stop(&mod);
  MODFILE_Free(&mod);
}

void a320_sound_modplay(unsigned int len, unsigned char *pMod) {
  unsigned int ret;
  
  ret = MODFILE_Set(pMod, len, &mod);

  mod.musicvolume = 64;
  mod.sfxvolume = 64;
  MODFILE_Start(&mod);
  /*
   * mod       - A pointer to the MODFILE structure of
   *             which the output format shall be changed
   * freq      - Output frequency. Common values are
   *             11025Hz, 22050Hz and 44100Hz
   * channels  - 1 for mono and 2 for stereo
   * bits      - 8 or 16 are valid values
   * mixsigned - TRUE if the output stream shall consist
   *             of signed values
  */ 
  MODFILE_SetFormat(&mod, A320_SAMPLE_RATE, 1, 16, 1);
  a320_sound_pausei = 0;
  a320_sound_mod_stop = 0;
}

void a320_sound_xmplay(unsigned int len, unsigned char *pMod) {
  unsigned int ret;
  
  ret = MODFILE_SetXM(pMod, len, &mod);
  //ret = MODFILE_Set(pMod, len, &mod);
  
  mod.musicvolume = 64;
  mod.sfxvolume = 64;
  MODFILE_Start(&mod);
  /*
   * mod       - A pointer to the MODFILE structure of
   *             which the output format shall be changed
   * freq      - Output frequency. Common values are
   *             11025Hz, 22050Hz and 44100Hz
   * channels  - 1 for mono and 2 for stereo
   * bits      - 8 or 16 are valid values
   * mixsigned - TRUE if the output stream shall consist
   *             of signed values
  */ 
  MODFILE_SetFormat(&mod, A320_SAMPLE_RATE, 1, 16, 1);
  a320_sound_pausei = 0;
  a320_sound_mod_stop = 0;
}

int a320_sound_modgetstat(void) 
{
  return MODFILE_getplaystatus(&mod);
}

int a320_sound_getofsstatus(void) 
{
  return MODFILE_getofsstatus(&mod);
}

void a320_sound_wavplay(void* buffer, unsigned int size, unsigned int repeat, unsigned int freq)
{
  int8_t chan;

  chan = 0;
  while ((a320_channel[chan].sndbuf != buffer || a320_channel[chan].loop == 0) &&
	 a320_channel[chan].loop != 0 &&	 chan < A320_MIXCHANNELS)
    chan++;
  if (chan == A320_MIXCHANNELS)
    chan = -1;

  if ((chan >= 0) ) {
    a320_channel[chan].loop = repeat;
    a320_channel[chan].sndbuf = buffer;
    a320_channel[chan].buf = buffer;
    a320_channel[chan].len = size;
  }
}

void a320_sound_wavstop(uint8_t chan) {
  //a320_sound_pausei = 1;
  a320_sound_thread_mute();  
}
